%**********************************************************
%   SUE, Single User Editor
%   derived from sundry JED modes
%   only for the GNU/LINUX CONSOLE on stand-alone x86 PCs
%   targets only one mode: no mode!
%**********************************************************

  _traceback = 1;

%--- required prototypes ----------------------------------

define dls_grep ()              ;
define dls_read_dir (dir)       ;
define disk_dirlist (thisdir)   ;
define go_paragraph_down ()     ;
define go_paragraph_up ()       ;
define block_cancel ()          ;
define display_help ()          ;

%----------------------------------------------------------


define shiftkey_check ()
%----------------------------------------------------------
%   Shift not pressed:  cancel an existing character block
%                       just moving the cursor
%
%   Shift pressed:      start a character block
%                       moving the cursor
%
%   To start/stop extending a block, movement must be with keys:
%   Left/Right/Up/Down, Home/End/CtrlHome/CtrlEnd, PageUp/PageDn,
%   CtrlLeft/CtrlRight.
%
%   This function is not assigned to any key,
%   it is used with cursor movements to extend a selection,
%   more precisely selection of a character stream
%----------------------------------------------------------
{
    %   stop here on rectangles and line blocks
    if (what_block) return;

    if (shiftkey_select)
    {
        %   shiftkey is on
        %   if there is no selection, start one

        ifnot(markp) call ("set_mark_cmd");
    }

    %   shiftkey is off, cancel stream if any
    else if (markp) pop_mark (0);
}


%**********************************************************
%   search and replace functions incl. regular expressions
%   normal search always case insensitive
%**********************************************************


define search_mark_nchars (n)
%----------------------------------------------------------
%   highlight a string of length 'n' found in search
%----------------------------------------------------------
{
    if (markp) toggle_block (0);

    %   match to be highlighted as region, status shows mark
    %   go_right and go_left force scrolling past border

    call ("set_mark_cmd");
    go_right (n);
    call("exchange");

    %   update_sans_update_hook disturbs the line mark
    %   update hook is used to shape cursor

    update (1);
    pop_mark(0);
    ungetkey(getkey);
}

define searchx (direction, pat)
%----------------------------------------------------------
%   called from search dialog, s_srch
%----------------------------------------------------------
{
    variable ffound;

    push_mark ();

    if (SEARCH_TYPE)
    {
        if(direction) ffound = re_fsearch(pat);
        else ffound = re_bsearch(pat);
    }
    else
    {
        if(direction) ffound = fsearch(pat);
        else ffound = bsearch(pat);
    }

    if (ffound)
    {
        if (SEARCH_TYPE)
            search_mark_nchars (ffound-1);
        else
            search_mark_nchars (ffound);
        LAST_SEARCH = pat;
    }
    else flush("Not found: " + pat);

    pop_mark (not(ffound));
}


define search_forward ()
%----------------------------------------------------------
%   Will ask for a string and search forward starting at
%   the cursor and disregarding case. Will stop at the
%   first hit between start position and end of buffer;
%   will stay put if nothing found.
%----------------------------------------------------------
{
    current_mini_cmd = _function_name;

    variable sstr=Null_String, prompt, ffound;

    prompt = "Search (forward) for:";

    sstr = read__mini (prompt, Null_String, Null_String, 0);

    ifnot (strlen(sstr)) return;

    if(is_substr (sstr, "\n"))
    {
        %   0 if '\n' is not in string
        flush("Search string must not include a line break.");
        return;
    }

    if(markp) toggle_block (0);
    push_spot();

    SEARCH_TYPE=0;
    %   simple (non regex) search is case-insensitive
    %   s. search.c

    ffound = fsearch (sstr);

    if (ffound)
    {
        search_mark_nchars (strlen(sstr));
        LAST_SEARCH = sstr;
    }
    else flush("Not found.");

    if(ffound)
    {
        push_mark();
        pop_spot ();
        pop_mark(1);
    }
    else pop_spot();
}

define search_again_forward ()
%----------------------------------------------------------
%   Will repeat last search from current position forward.
%   Repeated search will be of the same kind as last one,
%   simple or regular expression.
%----------------------------------------------------------
{
    if (LAST_SEARCH=="") return;
    %   empty string, nothing to do

    if(markp) toggle_block (0);

    go_right(1);

    variable r;
    if (SEARCH_TYPE) r = re_fsearch (LAST_SEARCH);
    else r = fsearch (LAST_SEARCH);

    if (r)
    {
        if (SEARCH_TYPE) search_mark_nchars (r-1);
        else search_mark_nchars (r);
    }
    else
    {
        go_left (1);
        flush("Not found.");
    }
}


define search_again_backward ()
%----------------------------------------------------------
%   Will repeat last search from current position
%   backward. Repeated search will be of the same kind as
%   last one, simple or regular expression.
%----------------------------------------------------------
{
    if (LAST_SEARCH=="") return;
    %   empty string, nothing to do

    if(markp) toggle_block (0);

    variable r;
    if (SEARCH_TYPE) r = re_bsearch (LAST_SEARCH);
    else r = bsearch (LAST_SEARCH);

    if (r)
    {
        if (SEARCH_TYPE) search_mark_nchars (r-1);
        else search_mark_nchars (r);
    }
    else flush("Not found.");
}


define replace_on_demand (pat, rep)
%----------------------------------------------------------
%   For both regex and non regex replacements
%   pat is the string or pattern to be replaced, never empty!
%   rep is the replacement for it, possibly empty,
%   and it is a literal string not a pattern.
%----------------------------------------------------------
{
    variable
        i = 0,
        query = 1,
        ch,
        pat_len,
        tmp,
        tmpstr,
        search_fun,
        xx = 0,
        yy = 0;

    variable prompt = "   " +
    "Tab=skip   |   Space=replace   |   Enter=replace all   |   Escape=cancel";

    if (SEARCH_TYPE)
    {
        %   regex
        %   if no match, 0 returned - otherwise 1+match_length
        search_fun  =   &re_fsearch;
    }

    else
    {
        %   simple (non regex)
        %   if no match, 0 returned - otherwise match length
        search_fun  =   &fsearch;
    }

    while (pat_len = @search_fun(pat), pat_len > 0)
    {
        ifnot (query)
        {
            %   replacing all at once from cursor downwards

            if(SEARCH_TYPE) --pat_len;
            loop(pat_len) del();
            if(strlen(rep)) insert(rep);

            i++;
            continue;
        }

        flush(prompt);
        search_mark_nchars(pat_len);

        %   if instance is at bottom of window, recenter
        if (window_info('r') == what_line)
            recenter (window_info('r') / 2);

        ch = toupper (getkey);

        %   SLang does not use 'break' on switch cases
        switch (ch)
        {
            case ' ':
            %   space pressed, replace match
            if(SEARCH_TYPE) --pat_len;
            loop(pat_len) del();
            if(strlen(rep)) insert(rep);
            i++;
        }

        {
            case '\t':

            %   tab pressed, skip to next match
            go_right(1);
        }

        {
            case '\n' or case '\r':

            %   enter pressed, replace all to the end
            query = 0;
            xx = what_column();
            yy = what_line();
            continue;
        }

        {
            case '\033':

            %   ESC pressed, cancel
            %   grey keys also produce an escape
            %   avoid inserting their string

            flush_input ();
            pop_mark (0);
            break;
        }

        {
            %   default

            flush_input ();
            continue;
        }
        %endswitch
    }

    if(not query)
    {
        goto_line(yy);
        goto_column(xx);
    }

    flush("Total replacements: " + string(i));
}


define replace_interactivly ()
%----------------------------------------------------------
%   Will ask for a string to be replaced and a string to
%   use as replacement (even empty). It will then proceed
%   interactively allowing for selective replacements.
%   Replacing all at once is possible (it will work
%   from the cursor position downwards).
%
%   The search for the string to be replaced is a simple
%   search (non regex) case-insensitive, the replacement
%   is inserted as is.
%----------------------------------------------------------
{
    if (is_readonly) return;

    current_mini_cmd = _function_name;

    variable
        prmpt  = Null_String,
        prompt = Null_String,
        rstr   = Null_String,
        sstr   = Null_String;

    SEARCH_TYPE = 0;
    prmpt = "String to replace:";

    sstr = read__mini (prmpt, Null_String, Null_String, 0);

    ifnot (strlen(sstr)) return;

    prompt = "Replacement:";

    rstr = read__mini (prompt, Null_String, sstr, 0);

    if (rstr == sstr) return;

    ifnot(strlen(rstr))
    {
        if(LAST_CHAR==13);
            %   ok
        else return;
    }

    replace_on_demand (sstr, rstr);
}


%**********************************************************
%   functions for macro recording and re-playing
%**********************************************************


define macro_start_or_stop ()
%----------------------------------------------------------
%   Toggles recording of a macro on or off
%----------------------------------------------------------
{
    if (DEFINING_MACRO) call("end_macro");
    else call("begin_macro");
}


define macro_repeat ()
%----------------------------------------------------------
%   Will prompt for a number and repeat macro execution
%   that many times in a row.
%
%   Use with caution: if the repeat factor is excessive
%   and the presumed conditions for the macro are not met,
%   junk and/or destruction may ensue.
%----------------------------------------------------------
{
    current_mini_cmd = _function_name;

    variable
        r=0,
        prompt="Repeat last macro how many times?";

    r = integer(read__mini(prompt, Null_String, " 1", 2));

    loop (r) call("execute_macro");
}


define macro_play ()
%----------------------------------------------------------
%   Execute last macro, i.e. repeat once
%----------------------------------------------------------
{
    call("execute_macro");
}


%**********************************************************
%   blocks copying, pasting, filling etc.
%**********************************************************


define block_cancel ()
%----------------------------------------------------------
%   Will cancel current selection, highlight will disappear
%----------------------------------------------------------
{
    if(is_visible_mark) toggle_block(0);
}



define block_toggle_rectangle ()
%----------------------------------------------------------
%   Starts a rectangle, if there is none.
%   Cancels a rectangle, if there is one.
%----------------------------------------------------------
{
    if (is_readonly) return;
    toggle_block (1);

    %   not explained yet
    set_buffer_undo (1);
}


define block_kill_rectangle ()
%----------------------------------------------------------
%   Will delete a rectangle
%----------------------------------------------------------
{
    %   if no column block, nothing to do
    ifnot (check_rectangle) return;

    %   global variables provided by check_rectangle ()
    variable
        beg_col   = W_COL,
        end_col   = E_COL,
        beg_line  = N_ROW,
        end_line  = S_ROW;

    %   do not delete rectangle with cut_rect
    %   to avoid polluting the clipboard
    %   del_region does not work well here
    %   so delete piecewise while looping

    goto_line(beg_line);
    goto_column(beg_col);
    variable rwidth = end_col - beg_col +1;

    loop (end_line - beg_line + 1)
    {
        goto_column (end_col+1);
        %   why? because rectangle may be ragged
        %   and a filling needed

        goto_column (beg_col);
        loop(rwidth) del();
        go_down (1);
    }

    goto_line (beg_line);
    goto_column (beg_col);
    %   cursor in NW corner

    toggle_block (1);
}


define block_fill_rectangle ()
%----------------------------------------------------------
%   Will fill the currently highlighted rectangle with a
%   string, as entered in the minibuffer.
%
%   If the string is too short, it will be repeated and
%   possibly truncated. If it is too long, it will be
%   truncated. This is done line by line.
%
%   To blank a rectangle, do not enter any string, just
%   press Enter.
%----------------------------------------------------------
{
    %   if no column block, nothing to do
    ifnot (check_rectangle)
    {
        flush ("No rectangle, nothing to do.");
        return;
    }

    variable
        str    = Null_String,
        tmpstr = Null_String,
        l,
        width,
        beg_col,
        beg_line,
        end_col,
        end_line;

    current_mini_cmd = _function_name;

    %   global variables provided by check_rectangle ()
    beg_col   = W_COL;
    end_col   = E_COL;
    beg_line  = N_ROW;
    end_line  = S_ROW;

    str = read__mini ("Fill rectangle with:", Null_String, Null_String, 2);

    if (LAST_CHAR != '\r') return;
    ifnot(strlen (str)) str = " ";
    %   empty reply will blank rectangle

    tmpstr = str;
    l = strlen (str);
    ifnot(l) return;

    width = end_col - beg_col + 1;
    %   width corresponds to cursor inside rectangle

    if (l > width) str = substr (str, 1, width);
    %   string too long, cut to width

    if (l < width)
    {
        loop (width/l-1) str = str + tmpstr;
        %   string too short
        %   concatenate to itself as far as possible

        l = strlen (str);
        str = str + substr (str, 1, width - l);
        %   add a substr to fill gap to width
    }

    %   do not delete rectangle with cut_rect
    %   to avoid polluting the clipboard
    %   del_region does not work well here
    %   so delete piecewise while inserting

    goto_line(beg_line);
    goto_column(beg_col);

    loop (end_line - beg_line + 1)
    {
        %   required at end of line
        %   otherwise chars from following line may be deleted
        goto_column (beg_col + width);
        goto_column (beg_col);

        insert (str);
        %   insert is an intrinsic

        loop(width) del();
        go_down (1);
    }
    ifnot (strlen(strtrim(str))) go_up(1);

    toggle_block (1);

    %   if filled with something, highlight again
    %   otherwise just blank
    %if (strlen(strtrim(str))) restore_rectangle();
    %else
    {
        toggle_block(0);
        goto_line(end_line);
        goto_column(end_col);
    }
}



define block_toggle_lines ()
%----------------------------------------------------------
%   if whole lines are selected, cancel the selection
%   otherwise start whole lines selection
%----------------------------------------------------------
{
    toggle_block (2);
}


define block_save ()
%----------------------------------------------------------
%   Save the currently highlighted block (of any kind) to
%   a disk file prompting for a file name in the
%   minibuffer.
%
%   All the lines touched by the block will be fully
%   saved.
%----------------------------------------------------------
{
    ifnot (markp)
    {
        flush ("No selection, nothing to do.");
        return;
    }
    current_mini_cmd = _function_name;
    save_selection();
}


define block_shift_left ()
%----------------------------------------------------------
%   Will shift block to the left.
%----------------------------------------------------------
{
    if(is_readonly) return;

    ifnot(markp)
    {
        flush("No selection, nothing to do.");
        return;
    }

    variable tmpblk = what_block ();

    if (tmpblk == 1)
    {
        shift_rectangle_left ();
        return;
    }

    if (tmpblk == 2) shift_linetext_left ();
}


define block_shift_right ()
%----------------------------------------------------------
%   Will shift block (whole lines, rectangles) to the right.
%----------------------------------------------------------
{
    if(is_readonly) return;

    ifnot(markp)
    {
        flush("No selection, nothing to do.");
        return;
    }

    variable tmpblk = what_block ();

    if (tmpblk == 0)
    {
        flush("Selection must be whole lines or rectangles.");
        return;
    }

    if (tmpblk == 1)
    {
        shift_rectangle_right ();
        return;
    }

    if (tmpblk == 2) shift_linetext_right ();
}


define block_all ()
%----------------------------------------------------------
%   Selects entire buffer.
%----------------------------------------------------------
{
    if (bobp and eobp) return;

    if (markp) toggle_block (0);

    bob ();
    %   to begin of buffer

    toggle_block (2);
    %   start selection here

    eob ();
    %   extend to end of buffer
}


define block_paste ()
%----------------------------------------------------------
%   Will insert content of clipboard.
%----------------------------------------------------------
{
    if (markp)
    {
        flush("No.");
        return;
    }
    %   no insertion if block active

    call ("yank");

    %   from pastebuffer (character stream)
    %   or rectangle buffer
    %   whichever was last filled with block_copy

    %   cursor positioned immediately after insertion
    %   lineblocks get a newline at begin and end of insertion
}


define block_copy ()
%----------------------------------------------------------
%   Copy selection to clipboard.
%   No selection, no action
%----------------------------------------------------------
{
    ifnot (markp) return;

    if (check_rectangle)
    {
        copy_rect ();
        return;
    }

    push_spot ();
    call ("copy_region");
    pop_spot ();

    %   copy_region ->      copy_region_cmd             intrin.c
    %   this is not the function used above

    %   copy_region ->          copy_to_pastebuffer     keymap.c
    %   copy_to_pastebuffer ->  copy_region_to_buffer   paste.c
    %   copy_region_to_buffer                           paste.c
}


define block_append ()
%----------------------------------------------------------
%   Copy to clipboard and append selection to scratch buffer
%   No selection, no action
%----------------------------------------------------------
{
    ifnot (markp) return;

    if (check_rectangle) copy_rect ();
    else
    {
        push_spot ();
        call ("copy_region");
        pop_spot ();
    }

    %   copy_region ->      copy_region_cmd             intrin.c
    %   this is not the function used above

    %   copy_region ->          copy_to_pastebuffer     keymap.c
    %   copy_to_pastebuffer ->  copy_region_to_buffer   paste.c
    %   copy_region_to_buffer                           paste.c

    %   append selection to scratch buffer
    variable tmpbf=whatbuf ();
    sw2buf ("*scratch*");
    eob ();
    newline ();
    block_paste ();
    sw2buf (tmpbf);
}


define block_cut ()
%----------------------------------------------------------
%   cut to clipboard
%----------------------------------------------------------
{
    if (check_rectangle)
    {
        cut_rect ();
        return;
    }

    variable thisblock = what_block ();
    call ("kill_region");
    if (thisblock==2) del();

    %   rectangle_buffer or default pastebuffer used
    %   whichever was filled last
}


define block_cut_append ()
%----------------------------------------------------------
%   cut to clipboard and also append to scratch buffer
%----------------------------------------------------------
{
    if (check_rectangle) cut_rect ();
    else
    {
        variable thisblock = what_block ();
        call ("kill_region");
        if (thisblock==2) del();
    }

    %   append the region to scratch buffer, if you so want
    variable tmpbf=whatbuf ();
    sw2buf ("*scratch*");
    eob ();
    newline ();
    block_paste ();
    sw2buf (tmpbf);

    %   rectangle_buffer or default pastebuffer used
    %   whichever was filled last
}


%**********************************************************
%   functions to get upper/lower/toggle case
%**********************************************************


define case_xform (how)
{
    () = check_region (0);
    variable f;

    switch (how)
        {
            case 'u' or case 'U':
            f = &strup;
            %   upper case
        }
        {
            case 'l' or case 'L':
            f = &strlow;
            %   lower case
        }
        {
            % default: do nothing
            return;
        }

    %   we want the region to be selected after the xform
    () = dupmark ();

    variable cc = what_column();
    variable reg = (@f)(bufsubstr ());

    %   one mark lost to bufsubstr
    () = dupmark();

    %   one mark lost to del_region
    del_region ();
    insert (reg);

    %   one mark survived, extend to start position
    goto_column(cc);
}


define case_in_rectangle (how)
{
    %   rect was just checked, borders are known

    variable
        l1 = N_ROW,
        l2 = S_ROW,
        c1 = W_COL,
        c2 = E_COL;

    %   adjust column, cursor is inside rectangle
    c2++;

    toggle_block(1);
    goto_line(l1);

    while (l2 >= what_line)
    {
        goto_column (c1);
        push_mark();
        goto_column(c2);
        case_xform (how);
        pop_mark(0);
        go_down(1);
    }

    restore_rectangle();
}


define case_generic (how)
%----------------------------------------------------------
%   if block selected, operate on block
%   otherwise operate on current character
%   argument 'how' can be
%   l or L      for lower case
%   u or U      for upper case
%----------------------------------------------------------
{
    if (is_readonly) return;

    if (check_rectangle)
    {
        case_in_rectangle (how);
        return;
    }

    if (markp)
    {
        case_xform (how);
        return;
    }

    %   no block, operate on character

    push_mark();
    go_right (1);
    case_xform (how);
    pop_mark(1);
}


define case_lower ()
%----------------------------------------------------------
%   Will turn a selection lower case
%----------------------------------------------------------
{
    case_generic ('l');
}


define case_upper ()
%----------------------------------------------------------
%   Will turn a selection upper case
%----------------------------------------------------------
{
    case_generic ('u');
}


define case_toggle_initial ()
%----------------------------------------------------------
%   Will reverse case of word initial.
%   Only letters are considered to be an initial.
%----------------------------------------------------------
{
    if (is_readonly) return;

    if (markp) return;

    push_spot ();
    variable ln0 = what_line ();
    variable pn0 = _get_point ();

    begin_of_word_right ();
    begin_of_word_left ();

    skip_chars ("^\\w");
    %   we are only interested in letters
    %   but we may skip past original position
    %   so take precautions

    if ((ln0 < what_line) or (pn0 < _get_point))
    {
        pop_spot ();
        return;
    }

    push_mark ();
    go_right (1);

    variable just1 = bufsubstr();
    if (just1 == strup(just1)) just1 = strlow(just1);
    else just1 = strup(just1);

    call ("backward_delete_char");
    insert(just1);
    pop_spot ();
}


%**********************************************************
%   functions to move the cursor
%**********************************************************

define go_tof ()
%----------------------------------------------------------
%   Move the cursor to begin of definition of the function
%   where the cursor is in.
%
%   It only searches for an opening curly brace in column 1,
%   so it is error-prone.
%----------------------------------------------------------
{
   () = bol_bsearch_char (curlyL);
}


define go_bof ()
%----------------------------------------------------------
%   Move the cursor to begin of definition of the function
%   where the cursor is in.
%
%   It only searches for a closing curly brace in column 1,
%   so it is error-prone.
%----------------------------------------------------------
{
   () = bol_fsearch_char (curlyR);
}


define go_to_file ()
%----------------------------------------------------------
%   line selection must be present
%   extract file name and line number from selection
%   load file, go to line
%   any good way to extract file?
%   we try colon as delimiter as in gcc messages
%   failing which space is used
%----------------------------------------------------------
{
    variable
        target,
        ffile   = Null_String,
        lline   = Null_String,
        ttmp    = Null_String,
        nslash  = 0,
        L       = 0;

    if (what_block==1) return;
    toggle_block(0);
    toggle_block (2);

    ttmp=strtrim(bufsubstr);
    ifnot(strlen(ttmp)) return;

    bol();

    if (0 == is_substr (ttmp, ":"))
    {
        %   no colon in string, use space as delimiter
        target = strchop (ttmp, ' ', 0);
    }
    else target = strchop (ttmp, ':', 0);

    %   char stream is chopped up in components
    %   delimited by a colon (as usual in gcc)
    %   if no colon present, space used as delimiter

    %   target[0]   path to file
    %   target[1]   line number

    ffile = strtrim(target[0]);
    L = length(target);
    if (L>1) lline = strtrim(target[1]);

    %   if there is no slash in ffile, take it as basename
    %   if there is a slash, it must be the first char

    nslash = is_substr (ffile, slash);

    if (nslash > 1)
    {
        %   selection has slashes, does not start with slash
        %   it may be just the basename, prefix pwd
        ffile = dircat (getcwd(), ffile);
    }

    if (1 != file_status (ffile))
    {
        write_scratch ("Path to file not correct:");
        newline();
        newline();
        write_scratch (ffile);
        return;
    }

    % load file

    find_file(ffile);

    ifnot (Integer_Type == _slang_guess_type (lline)) return;
    goto_line(integer(lline));
}


define go_to_matching_fence ()
%----------------------------------------------------------
%   If cursor on any of (){}[] jump to match, if existent.
%----------------------------------------------------------
{
    if (markp and what_block () != 2)
    {
        flush("No.");
        return;
    }
    %   no action if there is a rectangle or a stream
    %   but ok if it is a line block

    () = find_matching_fence ();
}


define go_page_up ()
%----------------------------------------------------------
%   If cursor is not at top of window, move it to the top
%   of window. From there move it up page by page.
%
%   The page depends on the windows the cursor is in,
%   normally the entire screen minus menubar, status line
%   and minibuffer.
%----------------------------------------------------------
{
    shiftkey_check();

    variable start_line = window_line ();
    %   cursor is on start_line

    variable current_column = what_column ();
    %   save column position

    goto_top_of_window ();
    %   move to top of window, if not already there

    if (window_line () == start_line)
    {
        %   cursor on top, paginate
        call ("page_up");
        goto_top_of_window ();
    }
    goto_column (current_column);
}


define go_page_down ()
%----------------------------------------------------------
%   If cursor is not at bottom of window, move it to
%   bottom of window. From there move it down page by
%   page.
%
%   The page depends on the windows the cursor is in,
%   normally the entire screen minus menubar, status line
%   and minibuffer.
%----------------------------------------------------------
{
    shiftkey_check();

    variable start_line = window_line ();
    %   cursor is on start_line

    variable current_column = what_column ();
    %   save column position

    goto_bottom_of_window ();
    %   move to bottom of window, if not already there

    if (window_line() == start_line)
    {
        %   cursor at bottom of window, paginate
        call ("page_down");
        goto_bottom_of_window ();
    }

    goto_column (current_column);
}


define go_pageup_select ()
%----------------------------------------------------------
%   Function extends a selection using PgUp
%----------------------------------------------------------
{
    if (check_rectangle) return;
    shiftkey_check();
    go_page_up ();
}


define go_pagedn_select ()
%----------------------------------------------------------
%   Function extends a selection using PgDn
%----------------------------------------------------------
{
    if (check_rectangle) return;
    shiftkey_check();
    go_page_down ();
}


define go__to ()
%----------------------------------------------------------
%  This is a command to pull down a menu from where you
%  can choose to go to line, column, matching fence (i.e.
%  bracket), previous position, quick mark, mid-window.
%  These options are all explained in the keys help by
%  highlighting them and pressing F1.
%----------------------------------------------------------
{
    if (check_rectangle) return;
    menu_select_menu ("Global.&Search.&Goto");
}


define go_to_line ()
%----------------------------------------------------------
%   Will go to a line specified by number
%----------------------------------------------------------
{
    if (check_rectangle) return;
    current_mini_cmd = _function_name;

    shiftkey_check();

    variable s = 0;
    s = integer (read__mini ("Goto line:", Null_String, Null_String, 2));

    ifnot(s) return;

    goto_line (s);
    %   cursor is at bol
}


define go_to_column ()
%----------------------------------------------------------
%   Will go to a column specified by number
%   may go behind the current end of line
%----------------------------------------------------------
{
    if (check_rectangle) return;
    current_mini_cmd = _function_name;

    shiftkey_check();

    variable s = 0;
    s = integer (read__mini ("Goto column:", Null_String, Null_String, 2));

    ifnot (s) return;

    eol();

    variable S = 100 + what_column();

    if (s <= S) goto_column (s);
    else
    {
        goto_column (S);
        flush("Max. allowed: line length plus 100.");
        %   spaces will be inserted if necessary
    }
}


define go_cursor_left ()
%----------------------------------------------------------
%   Move the cursor one column to the left. If at begin of
%   line, wrap around to the end of preceding line.
%----------------------------------------------------------
{
    shiftkey_check ();
    call ("previous_char_cmd");
}


define go_cursor_right ()
%----------------------------------------------------------
%   Move the cursor one column to the right. If at end of
%   line (any trailing spaces included), wrap around to
%   begin of following line.
%----------------------------------------------------------
{
    shiftkey_check ();
    call ("next_char_cmd");
}


define go_cursorleft_select ()
%----------------------------------------------------------
%   Will move the cursor one left and extend selection
%----------------------------------------------------------
{
    shiftkey_check ();
    call ("previous_char_cmd");
}


define go_cursorright_select ()
%----------------------------------------------------------
%   Will move the cursor one right and extend selection
%----------------------------------------------------------
{
    shiftkey_check ();
    call ("next_char_cmd");
}


define go_word_right ()
%----------------------------------------------------------
%   Will move cursor one word right, no return
%----------------------------------------------------------
{
    shiftkey_check ();
    begin_of_word_right ();
}


define go_word_left ()
%----------------------------------------------------------
%   Will move the cursor one word left, no return
%----------------------------------------------------------
{
    shiftkey_check ();
    begin_of_word_left ();
}


define go_wordright_select ()
%----------------------------------------------------------
%   Will move one word right and extend selection
%----------------------------------------------------------
{
    shiftkey_check ();
    begin_of_word_right ();
}


define go_wordleft_select ()
%----------------------------------------------------------
%   Will move one word left and extend selection
%----------------------------------------------------------
{
    shiftkey_check ();
    begin_of_word_left ();
}


define go_up_vertically ()
%----------------------------------------------------------
%   The cursor moves to previous line positioning itself
%   on the same column as in current line. If necessary,
%   the previous line will be padded with spaces. This,
%   however, does not apply to a read-only buffer where
%   the cursor is not visible.
%
%   When the buffer is editable, a line will be trimmed
%   before the cursor moves from it. In any event, the
%   modified flag will not change.
%----------------------------------------------------------
{
    shiftkey_check();
    if(LOCKSTEP)
    {
        scroll_down();
        otherwindow();
        scroll_down();
        otherwindow();
        return;
    }
    call("previous_line_cmd");
}

define go_down_vertically ()
%----------------------------------------------------------
%   The cursor moves to next line positioning itself on
%   the same column as in current line. If necessary, the
%   next line will be padded with spaces. This, however,
%   does not apply to a read-only buffer where the cursor
%   is not visible.
%
%   When the buffer is editable, a line may be trimmed
%   before the cursor moves from it. In any event, the
%   modified flag will not change.
%----------------------------------------------------------
{
    shiftkey_check();
    if(LOCKSTEP)
    {
        scroll_up();
        otherwindow();
        scroll_up();
        otherwindow();
        return;
    }
    call("next_line_cmd");
}


define go_up_select ()
%----------------------------------------------------------
%   Will move one up on same column and extend selection
%----------------------------------------------------------
{
    shiftkey_check();
    call("previous_line_cmd");
}


define go_down_select ()
%----------------------------------------------------------
%   Will move one down on same column and extend selection
%----------------------------------------------------------
{
    shiftkey_check();
    call("next_line_cmd");
}

define go_paragraph_up ()
%----------------------------------------------------------
%   Move back one paragraph
%----------------------------------------------------------
{
    if (what_block () != 0) return;
    %   stay put when lines/rectangles highlighted

    ifnot(up(1)) return;
    backward_paragraph();
    ifnot(down(1)) return;
    if(is_whiteline) go_up(1);
}


define go_paragraph_down ()
%----------------------------------------------------------
%   Move forward one paragraph
%----------------------------------------------------------
{
    if (what_block () != 0) return;
    %   stay put when lines/rectangles highlighted

    ifnot(down(1)) return;
    forward_paragraph();
    while(is_whiteline)
        ifnot(down(1)) return;
}


define go_begin_line ()
%----------------------------------------------------------
%   Will move to begin of line
%----------------------------------------------------------
{
    shiftkey_check();
    bol();
}


define go_end_line ()
%----------------------------------------------------------
%   Go to end of line but before trailing whitespace.
%   If no whitespace, cursor will rest on newline code
%   except possibly at end of buffer.
%----------------------------------------------------------
{
    shiftkey_check();
    eol();
    %bskip_chars (" \t");
}


define go_beginline_select ()
%----------------------------------------------------------
%   Will move to begin of line and extend selection
%----------------------------------------------------------
{
    shiftkey_check();
    bol();
}


define go_endline_select ()
%----------------------------------------------------------
%   go to eol but before trailing whitespace
%   cursor will rest on newline code or eob-code
%----------------------------------------------------------
{
    shiftkey_check();
    eol();
    %bskip_chars (" \t");
}


define go_begin_buffer ()
%----------------------------------------------------------
%   Will move to begin of buffer
%----------------------------------------------------------
{
    if (check_rectangle)
    {
        flush(r_reminder);
        return;
    }

    shiftkey_check();
    bob();
}


define go_end_buffer ()
%----------------------------------------------------------
%   Will move to end of buffer
%----------------------------------------------------------
{
    if (check_rectangle)
    {
        flush(r_reminder);
        return;
    }

    shiftkey_check();
    eob();
}


define go_beginbuffer_select ()
%----------------------------------------------------------
%   Will move to begin of buffer and extend selection
%----------------------------------------------------------
{
    if (check_rectangle)
    {
        flush(r_reminder);
        return;
    }

    shiftkey_check();
    bob();
}


define go_endbuffer_select ()
%----------------------------------------------------------
%   Will move to end of buffer and extend selection
%----------------------------------------------------------
{
    if (check_rectangle)
    {
        flush(r_reminder);
        return;
    }

    shiftkey_check();
    eob();
}


define ed_delete_word ()
%----------------------------------------------------------
%   If cursor is on word, delete this word, trim whitespace
%   to next word, position on next word.
%   If word is on whitespace, trim (without gluing words)
%----------------------------------------------------------
{
    if (markp)
    {
        flush(s_reminder);
        return;
    }
    %   block is selected, do nothing

    variable
        tmp     = 0,
        leading = 1,
        between = 2;

    trim();
    if (is_posixwhite) return;
    skip_black ();
    push_mark ();
    bskip_black ();
    tmp = what_column ();
    del_region ();

    if (is_lilywhite == leading)
    {
        push_mark();
        skip_chars(" \t");
        del_region();
    }
    else trim();

    if (is_lilywhite == between)
    {
        go_right (1);
        forever
        {
            if (what_column () == tmp) break;
            insert (" ");
        }
    }
}


define ed_delete_here ()
%----------------------------------------------------------
%   If there is a block of any kind, the block will be
%   deleted with the cursor positioned on topmost/leftmost
%   character.
%
%   If there is no block, the character at the cursor will
%   be deleted. If at end of line, the newline code will
%   be deleted and the current line joined with the
%   following one.
%----------------------------------------------------------
{
    %   if block selected, delete block
    %   otherwise delete character

    if (check_rectangle)
    {
        block_kill_rectangle ();
        return;
    }

    if (markp)
    {
        variable thisblock = what_block ();
        del_region ();
        if (thisblock==2) del ();
    }

    else
    {
        if (is_whiteline)
        {
            if (bolp) trim ();
        }
        del();
    }
}


define ed_delete_backward ()
%----------------------------------------------------------
%   If there is a block of any kind, nothing is done.
%
%   Otherwise, the character on the left will be deleted.
%   If at begin of line, the newline code will be deleted
%   and the current line joined with the preceding one.
%----------------------------------------------------------
{
    if (markp)
    {
        flush(s_reminder);
        return;
    }

    call ("backward_delete_char");
}


define ed_undelete ()
%----------------------------------------------------------
%   When deleting words, lines and regions, the deletions
%   are stored in memory and can be retrieved with this
%   function.
%
%   When undeleting, cursor does not jump back to the
%   location where something was deleted. The previous
%   deletion is inserted at the cursor wherever it is, the
%   cursor is positioned after it.
%----------------------------------------------------------
{
    if (is_readonly) return;
    if (markp) return;
    current_mini_cmd = _function_name;

    variable rply = "n";

    rply = strtrim (read__mini
    ("Undelete last deletion at cursor (y/n)?:", " n", " n", 2));

    ifnot(strlen(rply)) return;

    if (rply=="y" or rply=="Y")
        insert_from_kill_array(0);
}


define ed_untab ()
%----------------------------------------------------------
%   Replace hard tabs with spaces. You will be asked
%   tab width (say 8) and start of tab stops (say 1).
%
%   Replacement is either in selection (if any) or from
%   cursor onwards. Cursor will not move.
%----------------------------------------------------------
{
    if (is_readonly)
    {
        flush("Buffer is read-only.");
        return;
    }

    if (check_rectangle)
    {
        flush(r_reminder);
        return;
    }

    current_mini_cmd = _function_name;

    variable
        n           = 0,
        q           = 0,
        ttab        = 0,
        first_stop  = 0,
        s           = Null_String;

    s = read__mini
    ("Untab: how many spaces for a tab?", Null_String, Null_String, 2);

    ifnot (strlen (s)) return;

    ttab = integer (s);

    %   unsuitable entries
    if (ttab < 1) return;

    s = read__mini
    ("Count tabstops from 0 or 1?", Null_String, " 1", 2);

    ifnot (strlen (s)) return;

    first_stop = integer(s);

    if (first_stop != 1) first_stop = 0;

    variable
        lnz = markp(),
        lnc = -1 + lnz + what_line ();

    push_spot();
    () = check_region (0);
    pop_mark (1);

    while (fsearch("\t"))
    {
        if(lnz)
        {
            if(what_line > lnc) break;
        }

        ++q;
        del();
        n = - first_stop + (what_column() mod ttab);
        loop(ttab - n) insert_char(' ');
    }

    pop_spot();

    if(lnz)
        s = "Tabs replaced in selection: ";
    else
        s = "Tabs replaced past cursor: ";

    flush(s + string(q));
}


define ed_quick_untab ()
%----------------------------------------------------------
%   Replace hard tabs with spaces. Tab stops are 1, 5, 9,
%   13, ...
%
%   It acts only from cursor onwards, cursor will not move.
%   Replacement is not a fixed number of spaces for a tab,
%   it is a variable number to reach the next tab stop.
%   Tab stops are assumed here to be 8 cells, count from 1.
%----------------------------------------------------------
{
    if (is_readonly)
    {
        flush("Buffer is read-only.");
        return;
    }

    if (check_rectangle)
    {
        flush(r_reminder);
        return;
    }

    variable
        n,
        q=0,
        ttab=8,
        first_stop=1,
        s=Null_String;

    push_spot ();

    while (fsearch ("\t"))
    {
        ++q;
        del();
        n = - first_stop + (what_column() mod ttab);
        loop(ttab - n) insert_char(' ');
    }

    pop_spot();

    flush("Tabs replaced: " + string(q));
}

define ed_quick_4untab ()
%----------------------------------------------------------
%   Replace hard tabs with spaces. Tab stops are 1, 5, 9,
%   13, ...
%
%   It acts only from cursor onwards, cursor will not move.
%   Replacement is not a fixed number of spaces for a tab,
%   it is a variable number to reach the next tab stop.
%----------------------------------------------------------
{
    if (is_readonly)
    {
        flush("Buffer is read-only.");
        return;
    }

    if (check_rectangle)
    {
        flush(r_reminder);
        return;
    }

    variable
        n,
        q=0,
        ttab=4,
        first_stop=1,
        s=Null_String;

    push_spot ();

    while (fsearch ("\t"))
    {
        ++q;
        del();
        n = - first_stop + (what_column() mod ttab);
        loop(ttab - n) insert_char(' ');
    }

    pop_spot();

    flush("Tabs replaced: " + string(q));
}


define ed_softtab ()
%----------------------------------------------------------
%   If a lines or rectangle block is selected, then shift
%   selection right by one cell. If a character block is
%   selected, nothing is done.
%
%   If no selection present, then insert spaces up to the
%   next tab stop.
%
%   Tab stops are computed starting in column 1 and adding
%   TAB which can be set locally in a buffer.
%   For example, if TAB is set to 4 the tab stops will be at
%   1, 5, 9, 13, 17, ...
%
%   Hard tabs can only be inserted as a literal and will
%   not be expanded and will not give visual alignment,
%   s. quoted_insert and insert_code_for key.
%----------------------------------------------------------
{
    if (markp)
    {
        block_shift_right ();
        return;
    }

    %   if function called from menu, stops here
    if(LAST_CHAR==13 or LAST_CHAR > 32)
    {
        flush ("No selection, nothing to do.");
        return;
    }

    spaces_for_tabs ();
    %   tab key will insert spaces up to next tab position
}


define ed_backtab ()
%----------------------------------------------------------
%   Line or rectangle selected
%     shift selection right
%
%   No selection
%     jump to previous tab stop
%----------------------------------------------------------
{
    if (markp)
    {
        block_shift_left ();
        return;
    }

    backtab ();
    %   jump back to tab stops
}


define ed_wrap_lines ()
%----------------------------------------------------------
%   Wraps lines from current line to end of paragraph,
%   cursor will not advance. Lines are wrapped at the
%   column determined by the variable WRAP which is 60 in
%   the C source but can be set in the resource file. It
%   can also be changed on the fly from the command line.
%----------------------------------------------------------
{
    if (is_readonly) return;

    if (markp)
    {
        flush(s_reminder);
        return;
    }

    if(is_whiteline) return;

    %   format from current line downwards
    bol ();
    push_spot ();
    newline ();
    call ("format_paragraph");
    pop_spot ();
    del();
    %   cursor is at bol of line where it was before
}


define ed_wrap_and_jump ()
%----------------------------------------------------------
%   Do wrap, then jump to next paragraph
%----------------------------------------------------------
{
    if (is_readonly) return;

    if (markp)
    {
        flush(s_reminder);
        return;
    }

    if(is_whiteline) return;

    ed_wrap_lines();

    % cursor is at bol of line where it was before
    go_paragraph_down ();

    variable lm = window_info ('r') / 2;
    recenter (lm);
}


define ed_unwrap ()
%----------------------------------------------------------
%   join lines of a paragraph to a single line
%----------------------------------------------------------
{
    if(is_whiteline) return;
    backward_paragraph();
    go_down(1);

    forever
    {
        eol();
        insert(" ");
        ifnot(down(1))
        {
            del ();
            bol();
            return;
        }
        if(is_whiteline) return;

        call ("backward_delete_char");
    }
}

define ed_display (stringA, stringB)
%----------------------------------------------------------
%   displays text built on the fly
%   used for buffer properties
%----------------------------------------------------------
{
    set_buffer_modified_flag (0);
    set_readonly(1);
    set_status_line ("  Press any key to exit", 0);

    dummy_menu(stringB);
    bob();
    while(is_whiteline) go_down(1);
    update(1);

    flush_input(getkey ());

    % unconditional
    delbuf(whatbuf);
}


define ed_manage_readonly ()
%----------------------------------------------------------
%   Only for use from menu, read-only on/off
%----------------------------------------------------------
{
    toggle_readonly ();
}



define ed_bflags(bf_flags)
%----------------------------------------------------------
%   Edit buffer flags for file properties
%----------------------------------------------------------
{
    %   1024 0x400  add CR to end of lines when writing buffer to disk
    %   512  0x200  file associated with buffer is a binary file
    %   256  0x100  do not backup
    %   128  0x080  force save when exiting
    %   64   0x040  buffer buried
    %   32   0x020  undo enabled
    %   16   0x010  overwrite mode
    %   8    0x008  read only
    %   4    0x004  file on disk modified
    %   2    0x002  auto save mode
    %   1    0x001  buffer modified

    variable
        in_this = Null_String,
        spaces5 = "     ",
        L       = 11,
        fflag   = Integer_Type [L],
        llabel  = String_Type [3,L],
        i       = 0x400,
        j       = 0;

    llabel [0,*] =
    [
        "end-of-line",          %1024
        "",                     %512
        "",                     %256
        "save on exit",         %128
        "",                     %64
        "undo",                 %32
        "",                     %16
        "read-only",            %8
        "file on disk",         %4
        "autosave",             %2
        "modified"              %1
    ];

    llabel [1,*] =
    [
        %   flag set
        "DOS/Windows",          %1024
        "yes",                  %512
        "yes",                  %256
        "forced",               %128
        "yes",                  %64
        "yes",                  %32
        "yes",                  %16
        "yes",                  %8
        "changed",              %4
        "yes",                  %2
        "yes"                   %1
    ];

    llabel [2,*] =
    [
        %   flag not set
        "unix",                 %1024
        "no",                   %512
        "no",                   %256
        "prompt",               %128
        "no",                   %64
        "no",                   %32
        "no",                   %16
        "no",                   %8
        "not changed",          %4
        "no",                   %2
        "no"                    %1
    ];

    while (i)
    {
        fflag[j] = (bf_flags & i);
        i = i shr 1;
        j++;
    }

    for (i=0; i<L; i++)
    {
        ifnot(strlen(llabel[0,i])) continue;
        insert(spaces5);
        insert(llabel[0,i]);
        goto_column(21);

        if (fflag[i]) in_this=llabel[1,i];
        else in_this=llabel[2,i];

        insert(in_this);
        newline();
    }
}


define ed_properties()
%----------------------------------------------------------
%   Will display buffer and file properties
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    %(file, dir, name, flags) = getbuf_info ([ buf ])

    %   get file info before creating special buffer
    %   this intrinsic from userinfo.c pushes into the stack
    %   the name of owner and group for the current file
    get_owner_group();
    variable grouP = ();
    variable owneR = ();

    variable
        tmpbf    = Null_String,
        stringA  = " /on the fly/",
        stringB  = "&Properties",
        full_fn  = Null_String,
        thisfile = Null_String,
        thisdir  = Null_String,
        thisbf   = Null_String,
        modn     = Null_String,
        modf     = 0,
        fs_flag  = 0,
        fstatus  = 0,
        fmode    = 0,
        bf_flags = 0,
        bf_modif = Null_String,
        bf_ro    = Null_String,
        trim_yn  = "yes",
        spaces5  = "     ",
        spaces4  = "    ";

    tmpbf = whatbuf ();
    fs_flag = fs_readonly ();
    fmode = is_readonly();

    (thisfile, thisdir, thisbf, bf_flags) = getbuf_info ();
    (modn, modf) = what_mode ();
    ifnot(TRIMSAVE) trim_yn = "no";

    full_fn = thisdir + thisfile;
    fstatus = file_status (full_fn);

    %   open temp buffer
    delbuf(stringA);
    sw2buf (stringA);
    newline();

    insert(spaces4);
    insert("Buffer:");
    goto_column (21);
    insert(thisbf);
    newline();
    ed_bflags(bf_flags);

    insert(spaces5);
    insert("tab width");
    goto_column (21);
    insert(string(TAB));
    newline();

    insert(spaces5);
    insert("wrap column");
    goto_column (21);
    if (modn=="w" and modf==1)
        insert(string(WRAP));
    else
        insert("no wrap");
    newline();

    insert(spaces5);
    insert("trim saving");
    goto_column (21);
    insert(trim_yn);
    newline();
    newline();

    insert(spaces4);
    insert("Directory:");
    goto_column (21);
    insert(thisdir);
    newline();
    newline();

    insert(spaces4);
    insert("File on disk:");
    goto_column (21);
    if (1 == fstatus) insert(thisfile);

    if (fstatus==1)
    {
        variable st;
        st = stat_file (full_fn);

        newline();
        insert(spaces5);
        insert("permissions:");
        goto_column (21);
        %   convert mode to octal first
        insert(ed_permissions(sprintf("%o",st.st_mode)));

        newline();
        insert(spaces5);
        insert("owner:");
        goto_column (21);

        insert(string(owneR));

        newline();
        insert(spaces5);
        insert("group:");
        goto_column (21);
        insert(string(grouP));

        newline();
        insert(spaces5);
        insert("bytes:");
        goto_column (21);
        insert(string(st.st_size));

        newline();
        insert(spaces5);
        insert("modified:");
        goto_column (21);
        insert(strftime ("%Y-%m-%d %H:%M", localtime (st.st_mtime)));

        if (fs_flag > 0)
        {
            newline();
            insert(spaces5);
            insert("warning!");
            goto_column (21);
            insert("READ-ONLY file system");
        }

        else if (fmode)
        {
            newline();
            insert(spaces5);
            insert("mode:");
            goto_column (21);
            insert("read-only");
        }
    }
    newline();
    newline();

    insert(spaces4);
    insert("Editor Root:");
    goto_column (21);
    insert(JED_ROOT);

    newline();
    newline();

    insert(spaces4);
    insert("Stack depth:");
    goto_column (21);
    insert(string(_stkdepth));
    flush (strftime ("  %Y-%m-%d %H:%M", localtime (_time)));
    ed_display (stringA, stringB);

    sw2buf(tmpbf);
}


define ed_evalbuffer ()
%----------------------------------------------------------
%   Let the sl-interpreter evaluate the current buffer
%   errors will be notified in the minibuffer
%----------------------------------------------------------
{
    flush_input();

    variable
        tmpbf = whatbuf,
        ndx   = is_substr (tmpbf, ".sl");

    ifnot(ndx)
    {
        flush("Is this an SLang script? Give it a proper sl-extension.");
        return;
    }

    variable e;
    try (e)
    {
        () = evalfile (tmpbf,tmpbf);
    }

    catch AnyError:
    {
        goto_line(e.line);
        vmessage("Line %d, %s", e.line, e.descr);
    }

    %   no further code!
}


define ed_whatchar ()
%----------------------------------------------------------
%   show in minibuffer the code for the character at cursor
%   dec/hex/oct will be shown
%----------------------------------------------------------
{
   vmessage ("%s", count_chars);
}


define ed_ins_pathfile()
{
   () = getbuf_info ();
   pop ();
   insert(());
   insert(());
}


define ed_ins_datetime()
{
    variable timE = time();
    (timE,) = strreplace (timE, "  ", " ", strlen(timE));

    variable
        datetime = Null_String,
        ww     = extract_element (timE, 0, ' '),
        mm     = extract_element (timE, 1, ' '),
        dd     = extract_element (timE, 2, ' '),
        tt     = extract_element (timE, 3, ' '),
        yyyy   = extract_element (timE, 4, ' '),
        space  = " ",
        hyphen = "-";

    if(strlen(dd)==1) dd = "0"+dd;

    datetime
        = strcat (space,yyyy,hyphen,mm,hyphen,dd,space,tt,space,ww,space);

    insert(datetime);
}


define ed_undo2()
{
    if(markp)
    {
        toggle_block(0);
        return;
    }

    call("undo");
}


define ed_grep()
%----------------------------------------------------------
%   Will grep the directory of current buffer flat or
%   recursevely, suggesting as search string the selection
%   on current line or the current word.
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable
        thisdir = Null_String,
        s_strng = Null_String;

    %   use string selected on one line, if any

    if(markp==1 and check_rectangle)
    {
        variable to_k = E_COL+1;

        if (PULSE_C)
        {
            %   cursor is right of mark
            goto_column(W_COL);
        }

        toggle_block(0);
        push_mark();
        goto_column(to_k);
    }

    if (chk_stream==1)
    {
        %   If there is a selection on a single line
        %   use it for grep. Selection will be lost
        %   since it is assumed to be ad hoc.
        s_strng = bufsubstr();
    }

    else
    {
        flush("Select search string on current line (or switch to file listing).");
        return;
    }

    %   we know what to search for

    (,thisdir,,) = getbuf_info ();
    dls(thisdir);
    dls_grep(s_strng);
}


define ask_close_buffer ()
%----------------------------------------------------------
%   Close current buffer after confirmation
%   should unsaved changes be present.
%----------------------------------------------------------
{
    if(whatbuf=="*scratch*");
    else if(whatbuf=="*traceback*");
    else if(buffer_modified)
    {
        variable rply = Null_String;
        rply = strlow(read__mini("Save changes (y/n)?", "n", "n", 2));
        if (rply == "y") disk_save_buffer();
        else if (rply == "n");
        else return;
    }
    close_buffer();
}


define next_buffer ()
{
    switch_buffer(0);
}


define prev_buffer ()
{
    switch_buffer(1);
}


%**********************************************************
%   file operations on disk
%**********************************************************


define disk_read_and_insert ()
%----------------------------------------------------------
%   reads a file from disk and inserts it into current buffer
%   only called from menu
%----------------------------------------------------------
{
    %   avoid interference with blocks
    if (markp)
    {
        flush(s_reminder);
        return;
    }

    if (is_readonly)
    {
        flush("Buffer is read-only.");
        return;
    }
    current_mini_cmd = _function_name;

    variable
        dir  = Null_String,
        file = Null_String;

    %   ask user for file

    (,dir,,) = getbuf_info ();
    file = read__mini ("File to insert:", "//", dir, 1);

    ifnot (strlen(file)) return;

    %   clean up entry
    (dir, file) = parse_filename (file);
    variable fullpath = dircat (dir, file);

    %   if entry ok, insert file
    if (file_status (fullpath) == 1)
    {
        %   move cursor to begin of a line before inserting
        %   so the insertion will not cut existing lines
        eol();
        newline();

        insert_file (fullpath);
        InsertBuffer = Null_String;
        return;
    }

    %   if control here, entry failed
    %   use entered directory to display files to pick up

    (,,InsertBuffer,) = getbuf_info ();

    disk_dirlist(dir);
    flush("Not found! Choose from listing then press Enter.");
}


define disk_wopen ()
{
    variable
        fullpath = Null_String,
        dir      = Null_String,
        file     = Null_String;

    %   ask user for file
    flush("Tab completion available!");
    sleep(1);
    file = read_path_from_mini ("Open:");

    ifnot(strlen(file)) return;

    InsertBuffer = Null_String;

    %   clean up entry
    (dir, file) = parse_filename (file);
    fullpath = dircat (dir, file);

    %   if entry ok, load file
    if (file_status (fullpath) == 1)
    {
        if(bufferp(path_basename(fullpath)))
            flush("File already open, not re-loaded.");

        () = find_file (fullpath);
    }

    else flush("Not found!");
}


define disk_open ()
%----------------------------------------------------------
%   Will prompt for file to open,
%   listing files if name not correct
%   called both with kbd shortcut and from menu
%----------------------------------------------------------
{
    if(nwinz)
    {
        disk_wopen ();
        return;
    }

    current_mini_cmd = _function_name;

    variable
        fullpath = Null_String,
        dir      = Null_String,
        file     = Null_String;

    %   ask user for file
    file = read_path_from_mini ("Open:");

    ifnot(strlen(file)) return;

    InsertBuffer = Null_String;

    %   was entry a directory?
    if (file_status (file) == 2)
    {
        %   directory exists, open it

        variable A = listdir(file);
        variable B = listsubs(file);

        if(lngth(B))
        {
            buffer_keystring(" ");
            %   this will avoid pausing on next command

            dls (file);
            flush (" ");
            dls_read_dir (file,file);
        }

        else if(lngth(A))
            dls(file);

        else flush ("Directory empty.");

        return;
    }

    %   clean up entry
    (dir, file) = parse_filename (file);
    fullpath = dircat (dir, file);

    %   if entry ok, load file
    if (file_status (fullpath) == 1)
    {
        if(bufferp(path_basename(fullpath)))
            flush("File already open, not re-loaded.");
            %   this message is overwritten if file changed on disk

        else assoc_delete_key (bkm, path_basename(fullpath));

        () = find_file (fullpath);
        return;
    }

    %   if control here, entry failed
    flush("Not found! Pick it up from listing.");

    disk_dirlist(dir);
    %   go to listing
}


define disk_dirlist (thisdir)
%----------------------------------------------------------
%   Will display listing of current directory
%----------------------------------------------------------
{
    if (thisdir == Null_String)
        (,thisdir,,) = getbuf_info ();

    dls (thisdir);
}


define disk_treeview ()
%----------------------------------------------------------
%   Will display directory tree
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable thisdir;
    (,thisdir,,) = getbuf_info ();

    ifnot(lngth(listsubs (thisdir)))
    {
        dls (thisdir);
        flush("No sub-dirs.");
        return;
    }

    dtr ();
}


%----------------------------------------------------------
%   calculator:
%   expression from the minibuffer,
%   result both in minibuffer and scratch.
%----------------------------------------------------------

define calc_display_1line (entry)
{
    if ( typeof (entry) == String_Type )
        insert (entry);

    else if (typeof (entry) != Array_Type)
        %   the common oneline scalar calculations
        insert (string (entry));

    else
        vinsert ("Object type not supported!");
}

define calc_display_lines ()
{
    variable rr;

    if (_stkdepth () < 1)
    {
        rr = "Nothing to do.";
        insert (rr);
    }

    else
    {
        _stk_reverse (_stkdepth());
        while (_stkdepth ())
        {
            rr = ();
            calc_display_1line (rr);
            newline ();
        }
    }

    return (rr);
}

define calc()
{
    current_mini_cmd = _function_name;

    variable cprompt = "Enter calc expression:";
    variable xprx = Null_String;
    variable rsult;

    mS = 2;
    xprx = read__mini (cprompt, Null_String, Null_String, mS);

    ifnot (strlen(xprx)) return;

    sw2buf("*scratch*");
    eob();
    newline();
    newline();
    insert (xprx + "\n=  ");

    variable ffloat = "%.4f";
    set_float_format (ffloat);
    _pop_n (_stkdepth());
    eval (xprx);
    rsult = calc_display_lines ();
    flush(string(rsult));
}


%**********************************************************
%
%   KEYS
%
%   Each buffer has a keymap attached, normally the global
%   keymap in editing buffers or a custom copy of slim_kmap
%   in special buffers.
%
%   Keymaps dedicated to the modifiers (a_kmap, c_kmap, s_c_kmap,
%   s_kmap) cannot be attached to a buffer and should not be
%   customized on the fly.
%
%   The variables
%
%   SHIFKEY_PRESSED, CTRLKEY_PRESSED and ALTKEY_PRESSED
%
%   are available in sl-files. They are used sparsely and their
%   usage is not reflected in the keymaps. For SHIFKEY_PRESSED,
%   usage can be tracked down from the help menu.
%
%   The SLang library offers no support for modifiers
%   under Linux. They are being used here gaming the library.
%
%**********************************************************

definekey ("del_to_bol",                Key_Del,    a_kmap);
definekey ("display_doclist",           Key_F1,     a_kmap);
definekey ("ed_evalbuffer",             Key_F8,     a_kmap);
definekey ("execute_cmd",               "x",        a_kmap);
definekey ("exit_sue",                  Key_F4,     a_kmap);
definekey ("go_bof",                    "1",        a_kmap);
definekey ("go_tof",                    "0",        a_kmap);
definekey ("go_to_file",                KP_Enter,   a_kmap);
definekey ("insert_code_for_key",       Key_Ins,    a_kmap);
definekey ("next_buffer",               Key_Right,  a_kmap);
definekey ("occur_functions",           Key_F3,     a_kmap);
definekey ("pcp",                       Key_BS,     a_kmap);
definekey ("prev_buffer",               Key_Left,   a_kmap);
definekey ("scroll_down",               "+",        a_kmap);
definekey ("scroll_up",                 "-",        a_kmap);
definekey ("select_pulldown_edit",      "e",        a_kmap);
definekey ("select_pulldown_file",      "f",        a_kmap);
definekey ("select_pulldown_help",      "h",        a_kmap);
definekey ("select_pulldown_misc",      "m",        a_kmap);
definekey ("select_pulldown_search",    "s",        a_kmap);
definekey ("select_pulldown_view",      "v",        a_kmap);
definekey ("toggle_windows",            Key_Space,  a_kmap);
definekey ("toggle_crmode",             Key_End,    a_kmap);
definekey ("toggle_wrap",               Key_F7,     a_kmap);

definekey ("block_copy",                Key_Ins,    c_kmap);
definekey ("bm_next",                   Key_Tab,    c_kmap);
definekey ("del_to_eol",                Key_Del,    c_kmap);
definekey ("duplicate_line",            Key_F9,     c_kmap);
definekey ("ed_delete_word",            Key_BS,     c_kmap);
definekey ("ed_wrap_and_jump",          Key_F7,     c_kmap);
definekey ("go_begin_buffer",           Key_Home,   c_kmap);
definekey ("go_end_buffer",             Key_End,    c_kmap);
definekey ("go_paragraph_down",         Key_Down,   c_kmap);
definekey ("go_paragraph_up",           Key_Up,     c_kmap);
definekey ("go_word_left",              Key_Left,   c_kmap);
definekey ("go_word_right",             Key_Right,  c_kmap);
definekey ("macro_repeat",              Key_F8,     c_kmap);
definekey ("occur_simple",              Key_F3,     c_kmap);
definekey ("prghelp_keyword",           Key_F1,     c_kmap);

definekey ("bm_list",                   "^B",       s_c_kmap);
definekey ("block_append",              "^C",       s_c_kmap);
definekey ("ed_quick_4untab",           "^D",       s_c_kmap);
definekey ("s_srch",                    "^F",       s_c_kmap);
definekey ("block_cut_append",          "^X",       s_c_kmap);
definekey ("disk_save_preserve_mtime",  "^S",       s_c_kmap);
definekey ("go_endbuffer_select",       Key_End,    s_c_kmap);
definekey ("go_beginbuffer_select",     Key_Home,   s_c_kmap);
definekey ("go_wordleft_select",        Key_Left,   s_c_kmap);
definekey ("go_wordright_select",       Key_Right,  s_c_kmap);
definekey ("bm_prev",                   Key_Tab,    s_c_kmap);

definekey ("block_cut",                 Key_Del,    s_kmap);
definekey ("go_down_select",            Key_Down,   s_kmap);
definekey ("go_endline_select",         Key_End,    s_kmap);
definekey ("newline",                   Key_Enter,  s_kmap);
definekey ("menu_last_entry",           Key_F10,    s_kmap);
definekey ("search_again_backward",     Key_F3,     s_kmap);
definekey ("prev_buffer",               Key_F6,     s_kmap);
definekey ("set_wrap_column",           Key_F7,     s_kmap);
definekey ("macro_play",                Key_F8,     s_kmap);
definekey ("duplicate_floor_char",      Key_F9,     s_kmap);
definekey ("go_beginline_select",       Key_Home,   s_kmap);
definekey ("block_paste",               Key_Ins,    s_kmap);
definekey ("go_cursorleft_select",      Key_Left,   s_kmap);
definekey ("go_pagedn_select",          Key_PgDn,   s_kmap);
definekey ("go_pageup_select",          Key_PgUp,   s_kmap);
definekey ("go_cursorright_select",     Key_Right,  s_kmap);
definekey ("ed_backtab",                Key_Tab,    s_kmap);
definekey ("go_up_select",              Key_Up,     s_kmap);

setkey ("block_all",                    "^A");
setkey ("bm_set",                       "^B");
setkey ("ed_quick_untab",               "^D");
setkey ("block_copy",                   "^C");
setkey ("shell",                        "^E");
setkey ("search_forward",               "^F");
setkey ("go__to",                       "^G");
setkey ("replace_interactivly",         "^H");

setkey ("block_fill_rectangle",         "^KF");
setkey ("case_toggle_initial",          "^KI");
setkey ("case_toggle_initial",          "^K^I");
setkey ("case_lower",                   "^KL");
setkey ("t_sort",                       "^KO");
setkey ("ed_whatchar",                  "^KD");
setkey ("block_save",                   "^KS");
setkey ("case_upper",                   "^KU");
setkey ("vertical_trim",                "^KV");
setkey ("ed_delete_word",               "^K"+Key_BS);

setkey ("disk_new_file",                "^N");
setkey ("disk_open",                    "^O");
setkey ("redraw",                       "^R");
setkey ("disk_save_buffer",             "^S");
setkey ("ed_undelete",                  "^U");
setkey ("block_paste",                  "^V");
setkey ("ask_close_buffer",             "^W");
setkey ("block_cut",                    "^X");
setkey ("del_line",                     "^Y");
setkey ("ed_undo2",                     "^Z");
setkey ("otherwindow",                  "^@");

setkey ("select_pulldown_help",         Key_F1);
setkey ("go_to_matching_fence",         Key_F2);
setkey ("search_again_forward",         Key_F3);
setkey ("disk_treeview",                Key_F4);
setkey ("reset_screen",                 Key_F5);
setkey ("next_buffer",                  Key_F6);
setkey ("ed_wrap_lines",                Key_F7);
setkey ("macro_start_or_stop",          Key_F8);
setkey ("duplicate_ceiling_char",       Key_F9);
setkey ("block_toggle_rectangle",       Key_F11);
setkey ("block_toggle_lines",           Key_F12);
setkey ("block_cancel",                 Key_Escape);

setkey ("ed_delete_backward",           Key_BS);
setkey ("ed_delete_here",               Key_Del);
setkey ("go_down_vertically",           Key_Down);
setkey ("go_end_line",                  Key_End);
setkey ("go_begin_line",                Key_Home);
setkey ("toggle_overwrite",             Key_Ins);
setkey ("go_cursor_left",               Key_Left);
setkey ("go_page_down",                 Key_PgDn);
setkey ("go_page_up",                   Key_PgUp);
setkey ("go_cursor_right",              Key_Right);
setkey ("ed_softtab",                   Key_Tab);
setkey ("go_up_vertically",             Key_Up);

setkey ("select_menubar",               Key_F10);
%   F10 works as a toggle: it is defined in menu.c
%   to exit the menus and return to editing buffer


%%%%%%%%%%  TEST ZONE   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define format_keymap()
{
    replace(string(char(9)), " ");
    bob();
    replace("plain ", "");
    bob();
    variable cc;
    %%variable cA = 16;
    variable cA = 24;
    variable cB = cA+13;

    while (fsearch ("="))
    {
        cc = what_column();

        if (cc < cB) cc = cB - cc;
        else cc=1;

        loop (cc) insert_char (' ');
        eol();
    }
    bob();

    while (fsearch (" keycode"))
    {
        cc = what_column();

        if (cc < cA) cc = cA     - cc;
        else cc=0;

        loop (cc) insert_char (' ');
        eol();
    }

    bob();

    while (fsearch ("="))
    {
        cc = what_column();

        if (cc > cB) cc = cc-cB;
        else cc=0;

        loop (cc) call ("backward_delete_char");
        eol();
    }
}


define ttest()
%----------------------------------------------------------
%   Does nothing specific, used to test user code.
%   Not same as test() which is an intrinsic
%   used to test C code from intrin.c
%----------------------------------------------------------
{

}
setkey ("ttest", "^t");
